#!/usr/bin/env python3
# coding: utf-8

import datetime

from pymongo.collection import Collection, ReturnDocument
from pymongo.errors import DuplicateKeyError


class SerialNumber:
    __slots__ = ['prefix', 'number', 'length']

    @staticmethod
    def _get_collection() -> Collection:
        raise NotImplementedError

    def __init__(self, prefix='ID', length=6):
        coll = self._get_collection()
        doc = coll.find_one_and_update(
            {'_id': prefix}, {'$inc': {'i': 1}},
            upsert=True, return_document=ReturnDocument.AFTER,
        )
        self.prefix = prefix
        self.number = doc['i']
        self.length = length

    def __str__(self):
        return self.prefix + str(self.number).zfill(self.length)


class NamedLock(object):
    @staticmethod
    def _get_collection() -> Collection:
        raise NotImplementedError

    def __init__(self, name: str, ttl=12):
        self.name = name
        self.coll = self._get_collection()
        self.ttl = ttl

    def acquire(self):
        now = datetime.datetime.now()
        expire_at = now + datetime.timedelta(seconds=self.ttl)
        self.coll.delete_many({'expire_at': {'$lt': now}})
        record = {'_id': self.name, 'expire_at': expire_at}
        try:
            self.coll.insert_one(record)
        except DuplicateKeyError:
            return False
        return True

    def release(self):
        self.coll.delete_one({'_id': self.name})



#!/usr/bin/env python3
# coding: utf-8

from collections import defaultdict
from typing import Union

from bson import ObjectId
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.database import Database

from joker.mongodb import utils

NoneType = type(None)


class DatabaseWrapper:
    def __init__(self, db: Database):
        self.db = db

    def inspect_storage_sizes(self):
        return utils.inspect_mongo_storage_sizes(self.db)

    def print_storage_sizes(self):
        return utils.print_mongo_storage_sizes(self.db)


class CollectionWrapper:
    def __init__(self, coll: Collection, filtr=None, projection=None):
        self.coll = coll
        self.filtr = filtr or {}
        self.projection = projection

    def exist(self, filtr: Union[ObjectId, dict]):
        return self.coll.find_one(filtr, projection=[])

    def find_recent_by_count(self, count=50) -> Cursor:
        cursor = self.coll.find(self.filtr, projection=self.projection)
        return cursor.sort([('_id', -1)]).limit(count)

    def find_most_recent_one(self) -> dict:
        recs = list(self.find_recent_by_count(1))
        if recs:
            return recs[0]

    def _insert(self, records):
        if records:
            self.coll.insert_many(records, ordered=False)

    @staticmethod
    def _check_for_uniqueness(records, uk):
        vals = [r.get(uk) for r in records]
        uniq_vals = set(vals)
        if len(vals) != len(uniq_vals):
            raise ValueError('records contain duplicating keys')

    def make_fusion_record(self):
        fusion_record = {}
        contiguous_stale_count = -1
        for skip in range(1000):
            record = self.coll.find_one(sort=[('$natural', -1)], skip=skip)
            if not record:
                continue
            contiguous_stale_count += 1
            for key, val in record.items():
                if not record.get(key):
                    fusion_record[key] = val
                    contiguous_stale_count = -1
            if contiguous_stale_count > 10:
                return fusion_record
        return fusion_record

    def query_uniq_values(self, fields: list, limit=1000):
        latest = [('_id', -1)]
        records = self.coll.find(sort=latest, projection=fields, limit=limit)
        uniq = defaultdict(set)
        for key in fields:
            for rec in records:
                val = rec.get(key)
                uniq[key].add(val)
        return uniq



#!/usr/bin/env python3
# coding: utf-8

from gridfs import GridFS
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database
from volkanic.compat import cached_property, abstract_property

from joker.mongodb import utils
from joker.mongodb.wrappers import DatabaseWrapper, CollectionWrapper


class MongoClientExtended(MongoClient):
    """An extended client-side representation of a mongodb cluster."""
    
    get_db = MongoClient.get_database

    def __repr__(self):
        cn = self.__class__.__name__
        return "{}({})".format(cn, self._repr_helper())

    def inspect_storage_sizes(self):
        return utils.inspect_mongo_storage_sizes(self)

    def print_storage_sizes(self):
        return utils.print_mongo_storage_sizes(self)

    def get_dbw(self, db_name: str) -> DatabaseWrapper:
        return DatabaseWrapper(self.get_database(db_name))

    def get_coll(self, db_name: str, coll_name: str) -> Collection:
        db = self.get_database(db_name)
        return db.get_collection(coll_name)

    def get_collw(self, db_name: str, coll_name: str) -> CollectionWrapper:
        coll = self.get_coll(db_name, coll_name)
        return CollectionWrapper(coll)

    def get_gridfs(self, db_name: str, coll_name: str = 'fs') \
            -> GridFS:
        # avoid names like "images.files.files"
        if coll_name.endswith('.files') or coll_name.endswith('.chunks'):
            coll_name = coll_name.rsplit('.', 1)[0]
        db = self.get_database(db_name)
        return GridFS(db, collection=coll_name)


class MongoInterface:
    """A interface for multiple mongodb clusters."""

    def __init__(self, hosts: dict = None, default: list = None):
        default = default or ['lh', 'test']
        assert len(default) == 2
        if hosts is None:
            hosts = {}
        hosts.setdefault('lh', {})
        self._clients = {}
        self.default = default
        self.hosts = hosts

    def get_mongo(self, host: str = None) -> MongoClientExtended:
        if host is None:
            host = self.default[0]
        try:
            return self._clients[host]
        except KeyError:
            params = self.hosts.get(host, {'host': host})
            mongo = MongoClientExtended(**params)
            self._clients[host] = mongo
            return mongo

    @property
    def db(self) -> Database:
        return self.get_db(*self.default)

    @property
    def dbw(self) -> DatabaseWrapper:
        return self.get_dbw(*self.default)

    def get_db(self, host: str, db_name: str) -> Database:
        mongo = self.get_mongo(host)
        return mongo.get_db(db_name)

    def get_dbw(self, host: str, db_name: str) -> DatabaseWrapper:
        mongo = self.get_mongo(host)
        return mongo.get_dbw(db_name)

    def get_coll(self, host: str, db_name: str, coll_name: str) \
            -> Collection:
        mongo = self.get_mongo(host)
        return mongo.get_coll(db_name, coll_name)

    def get_collw(self, host: str, db_name: str, coll_name: str) \
            -> CollectionWrapper:
        mongo = self.get_mongo(host)
        return mongo.get_collw(db_name, coll_name)

    def get_gridfs(self, host: str, db_name: str, coll_name: str = 'fs') \
            -> GridFS:
        mongo = self.get_mongo(host)
        return mongo.get_gridfs(db_name, coll_name)


class GIMixinMongo:
    @abstract_property
    def conf(self) -> dict:
        return NotImplemented

    @cached_property
    def mongo(self) -> MongoClientExtended:
        params = self.conf.get('mongo', {})
        return MongoClientExtended(**params)


class GIMixinMongoi:
    @abstract_property
    def conf(self) -> dict:
        return NotImplemented

    @cached_property
    def db(self) -> Database:
        return self.mongoi.db

    @cached_property
    def dbw(self) -> DatabaseWrapper:
        return self.mongoi.dbw

    @cached_property
    def mongo(self) -> MongoClientExtended:
        return self.mongoi.get_mongo()

    @cached_property
    def mongoi(self) -> MongoInterface:
        return MongoInterface(
            self.conf.get('mongoi'),
            self.conf.get('mongoi-default'),
        )



#!/usr/bin/env python3
# coding: utf-8

import datetime

from bson import ObjectId


def in_(vals):
    return {'$in': vals}


def exclude(keys):
    return {k: False for k in keys}


def exists(*keys):
    return {k: {'$exists': True} for k in keys}


def oid_filter_by_datetime(
        start: datetime.datetime = None,
        end: datetime.datetime = None) -> dict:
    filtr = {}
    if start is not None:
        filtr['$gt'] = ObjectId.from_datetime(start)
    if end:
        filtr['$lt'] = ObjectId.from_datetime(end)
    return filtr


def oid_filter_recent(days=30, seconds=0):
    delta = datetime.timedelta(days=days, seconds=seconds)
    start = datetime.datetime.now() - delta
    return oid_filter_by_datetime(start, None)



#!/usr/bin/env python3
# coding: utf-8

import functools
import sys
import traceback

from pymongo.collection import Collection


def _show_call_stack(func):
    @functools.wraps(func)
    def ret_func(*args, **kwargs):
        print(func, file=sys.stderr)
        rv = func(*args, **kwargs)
        traceback.print_stack()
        return rv

    return ret_func



def patch_mongo_for_debug(wrapper_func=None):
    if wrapper_func is None:
        wrapper_func = _show_call_stack
    Collection.find = wrapper_func(Collection.find)
    Collection.find_one = wrapper_func(Collection.find_one)
    Collection.update = wrapper_func(Collection.update)
    Collection.update_one = wrapper_func(Collection.update_one)
    Collection.update_many = wrapper_func(Collection.update_many)


#!/usr/bin/env python3
# coding: utf-8

# from setup.py.jinja2

import os
import re

import setuptools
from setuptools import find_namespace_packages

# CAUTION:
# Do NOT import your package from your setup.py

_nsp = 'joker'
_pkg = 'mongodb'
_names = [_nsp, _pkg]
_names = [s for s in _names if s]


def read(filename):
    with open(filename) as f:
        return f.read()


def _find_version():
    names = _names + ['__init__.py']
    path = os.path.join(*names)
    root = os.path.dirname(__file__)
    path = os.path.join(root, path)
    regex = re.compile(
        r'''^__version__\s*=\s*('|"|'{3}|"{3})([.\w]+)\1\s*(#|$)''')
    with open(path) as fin:
        for line in fin:
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            mat = regex.match(line)
            if mat:
                return mat.groups()[1]
    raise ValueError('__version__ definition not found')


config = {
    'name': 'joker-mongodb',
    'version': _find_version(),
    'description': 'access mongodb with handy utilities and fun',
    'keywords': '',
    'url': 'https://github.com/frozflame/joker-mongodb',
    'author': 'frozflame',
    'author_email': 'frozflame@outlook.com',
    'license': "GNU General Public License (GPL)",
    'packages': find_namespace_packages(include=['joker.*']),
    'zip_safe': False,
    'install_requires': read("requirements.txt"),
    'classifiers': [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    # ensure copy static file to runtime directory
    'include_package_data': True,
    'long_description': read('README.md'),
    'long_description_content_type': "text/markdown",
}

if _nsp:
    config['namespace_packages'] = [_nsp]

setuptools.setup(**config)



#!/usr/bin/env python3
# coding: utf-8
from typing import Union

from joker.cast.numeric import human_filesize
from joker.textmanip.tabular import tabular_format

from pymongo import MongoClient
from pymongo.database import Database


def inspect_mongo_storage_sizes(target: Union[MongoClient, Database]):
    if isinstance(target, MongoClient):
        return {r['name']: r['sizeOnDisk'] for r in target.list_databases()}
    size_of_collections = {}
    for coll_name in target.list_collection_names():
        info = target.command('collStats', coll_name)
        size_of_collections[info['ns']] = info['storageSize']
    return size_of_collections


def print_mongo_storage_sizes(target: Union[MongoClient, Database]):
    s_rows = list(inspect_mongo_storage_sizes(target).items())
    s_rows.sort(key=lambda r: r[1], reverse=True)
    rows = []
    for k, v in s_rows:
        num, unit = human_filesize(v)
        rows.append([round(num), unit, k])
    for row in tabular_format(rows):
        print(*row)